"""
tradegame provides the EliteGame class for accessing game journal data for Elite: Dangerous.

This module provides access to current game state by reading Elite: Dangerous
journal files and real-time status data. It extracts trade-relevant information
such as current location, ship configuration, cargo capacity, and credits.

If the journal is not in a standard place, such as on Linux or SteamDeck,
you can set the environment variable "ELITE_JOURNAL_PATH".

Journal File Format:
    Line-delimited JSON files written to OS-specific saved games directory.
    Each line is a complete JSON object representing a game event.

Supported Data:
    - Commander name and credits
    - Current system and station (if docked)
    - Ship type and configuration
    - Cargo capacity and current cargo mass
    - Jump range (laden and unladen)

Example:
    # Default everything.
    from tradedangerous.tradegame import EliteGame
    
    game = EliteGame()
    print(f"Commander: {game.commander_name}")
    
    
Example:
    # Builder-pattern with "TradeEnv" for arguments.
    from tradedangerous import TradeEnv
    from tradedangerous.tradegame import EliteGame
    
    tdenv = TradeEnv(debug=2, path="/Elite/Saved Games")
    elite = EliteGame(tdenv=tdenv)
    print(f"Current Station: {elite.current_station}")


Example:
    # Control which files are loaded; the same argument names can be
    # set as properties on a tdenv instance instead.
    from tradedangerous.tradegame import EliteGame, JournalLoad as jl, JsonFiles as jf
    elite = EliteGame(
                path="/Elite/Saved Games",
                # journals alternatives: None (don't load), or an integer limit,
                journal_load=jl.ALL_JOURNALS,
                extra_jsons=[jf.Outfitting],
    )
"""
from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
import datetime
import enum
import os
import re
import time
import typing

import orjson

from .tradeenv import TradeEnv
from .tradeexcept import TradeException

if typing.TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from typing import Any, Final


# Environment variable to override
ELITE_JOURNAL_ENVVAR = "ELITE_JOURNAL_PATH"


class GameFolderError(TradeException):
    """ GameFolderError indicates we could not find where the saved game data is. """
    def __init__(self, msg: str) -> None:
        super().__init__(
            "Journal functionality needs to know where Elite's save-game data "
            f"is stored, aka the Journal Folder: {msg}. "
            f"Set the '{ELITE_JOURNAL_ENVVAR}' environment variable to the correct path for your machine"
        )


class GameException(TradeException):
    """ GameException is base exception raised when a game file cannot be loaded or parsed."""


class GameDataException(TradeException):
    def __str__(self) -> str:
        return (
            super().__str__() + "\n" +
            "- If you just launched the game, try again after the game has loaded.\n"
            "- If the game is already running, you may need to change screen/dock/undock "
                "for the game to update/generate the file\n"
            f"- Check the journal folder is correct and/or set the {ELITE_JOURNAL_ENVVAR} environment variable\n"
            "- This can happen if the game updates the file just as you run the command; try again"
        )


class JournalException(GameDataException):
    """ JournalException is raised when a journal file cannot be loaded or parsed. """


class JsonException(GameDataException):
    """ JsonException is raised when a JSON file cannot be loaded or parsed. """


class MissingJsonError(JsonException):
    """ MissingJsonError indicates a required JSON file was not present. """
    def __init__(self, msg: str) -> None:
        super().__init__(f"Required save-game json data file was missing: {msg}.\n")


@dataclass
class Status:
    timestamp: str | None = None
    commander: str | None = None
    fid: str | None = None
    star_system: str | None = None
    system_address: int | None = None
    station_name: str | None = None
    station_type: str | None = None
    docked: bool = False
    landed: bool = False
    credits: int | None = None
    destination: dict[str, Any] | None = None
    cargo_space: int | None = None
    cargo_load: int | None = None


@enum.unique
class JournalLoad(enum.IntEnum):
    DO_NOT_LOAD  = 0
    MOST_RECENT  = 1
    ALL_JOURNALS = -1


@enum.unique
class JsonFiles(str, enum.Enum):
    """Enumeration of supported JSON data files."""
    BACKPACK    = "Backpack.json"
    CARGO       = "Cargo.json"
    MARKET      = "Market.json"
    MODULESINFO = "ModulesInfo.json"
    NAVROUTE    = "NavRoute.json"
    OUTFITTING  = "Outfitting.json"
    SHIPLOCKER  = "ShipLocker.json"
    SHIPYARD    = "Shipyard.json"
    STATUS      = "Status.json"


# The format used for timestamps.
TS_FMT = "%Y-%m-%dT%H:%M:%SZ"


# Default journal directory paths by OS
WINDOWS_JOURNAL_DIR: Final[Path]     = Path(Path.home(), "Saved Games", "Frontier Developments", "Elite Dangerous")
MACOS_JOURNAL_DIR: Final[Path]       = Path(Path.home(), "Library", "Application Support", "Frontier Developments", "Elite Dangerous")


@contextmanager
def bench(tdenv: TradeEnv, label: str) -> Generator[None]:
    """ Context manager bench for measuring elapsed time of code blocks. """
    start = time.perf_counter()
    yield
    end = time.perf_counter()
    elapsed = end - start
    tdenv.DEBUG0(f"{label} took {elapsed:.6f} seconds")


class EliteGame:
    """
    EliteGame class reads the various data files Frontier expose for Elite: Dangerous tools
    to access current game state: the journal files and a collection of json files.
    
    By default reads the most recent journal file and DEFAULT_JSON_FILES. Additional files
    can be loaded.
    
    Reads the most recent journal file to extract current game state including
    location, ship configuration, cargo, and credits. Supports both Windows
    and macOS journal locations.

    Interpretation of the data is mostly left to the caller. Some common data can be
    summarized for you by calling get_status().
    
    Example:
        game = EliteGame()
        status = game.get_status()
        if status.docked:
            print(f"Docked at {status.star_system}/{status.station_name}"
        else:
            print(f"In space at {status.star_system}")
    """
    # Class variables/Constants.
    DEFAULT_JOURNAL_LOAD: Final[JournalLoad] = JournalLoad.MOST_RECENT
    DEFAULT_JSON_FILES: Final[list[JsonFiles]] = [
        JsonFiles.STATUS,       # Fairly stock data
        JsonFiles.MODULESINFO,  # So we can determine your cargo capacity
    ]
    
    # Member variables.
    tdenv: TradeEnv                             # context object we're associated with
    path: Path                                  # Elite's save-game folder
    journal_load: JournalLoad                   # journal files we want to load
    json_files: list[JsonFiles]                 # json files we want to load
    
    # Loaded data.
    journals_loaded: list[Path]                 # json files we've loaded or empty
    journal: list[dict[str, dict]]              # events in load order (oldest->newest)
    history:  dict[str, list[dict[str, dict]]]  # oldest-first history per event name
    events: dict[str, dict[str, Any]]           # most-recent only instance of each event
    
    jsons_loaded: list[JsonFiles]               # files actually loaded
    json_data: dict[JsonFiles, dict[str, Any]]  # json file -> the data we loaded
    
    loaded_time: datetime.datetime              # when the load was completed
    
    status: Status | None                       # cache results from get_status()
    
    def __init__(
        self,
        *,
        tdenv: TradeEnv | None = None,
        path: str | Path | None = None,              # platform specific default
        journal_load: JournalLoad | None = None,     # default: EliteGame.DEFAULT_JOURNAL_LOAD
        json_files: list[JsonFiles] | None = None,   # default: EliteGame.DEFAULT_JSON_FILES
        extra_jsons: list[JsonFiles] | None = None,  # default: []
    ) -> None:
        """
            Initialize EliteGame instance by reading journal and JSON files.
            
            :param tdenv:        Optional TradeEnv instance for parameters, configuration, logging, etc.
                                 If path, journal_load, and/or json_files are not provided,
                                 tdenv will be used for defaults before falling back to the class defaults.
            :param path:         Optional path to Elite: Dangerous saved games directory.
                                 If not provided, uses tdenv.journal_path or discovers default based on OS.
            :param journal_load: Optional JournalLoad enum to control if/how many journals
                                 are loaded. If an integer value is provided, it indicates the
                                 maximum number of journals (most recent first) to load.
                                 If None, uses tdenv or class default.
            :param json_files:   Optional list of JsonFiles enum values indicating
                                   which JSON files to load. If None, uses tdenv or class default.
        """
        self.tdenv = tdenv or TradeEnv()
        self.journals_loaded = []
        self.journal = []
        self.history = {}
        self.events = {}
        self.jsons_loaded = []
        self.json_data = {}
        self.status = None
        
        # Populate the options in reverse order of priority: default, tdenv, explicit parameter.
        self.path = journal_path(self.tdenv, path)
        self.tdenv.DEBUG0("elite journal path: {}", self.path)
        
        self.journal_load = self.DEFAULT_JOURNAL_LOAD
        if tdenv and (env_journal_load := getattr(tdenv, "journal_load", None)):
            self.journal_load = env_journal_load
        if journal_load is not None:
            self.journal_load = journal_load
        self.tdenv.DEBUG1("elite game journal load: {!r}", self.journal_load)
        
        self.json_files = self.DEFAULT_JSON_FILES.copy()
        if tdenv and (env_json_files := getattr(tdenv, "journal_json_files", None)) is not None:
            self.json_files = env_json_files
        if json_files is not None:  # explicit None check so that passing `json_files=[]` works.
            self.json_files = json_files
        
        if tdenv and (env_extra_jsons := getattr(tdenv, "journal_extra_jsons", None)) is not None:
            self.json_files.extend(env_extra_jsons)
        if extra_jsons:
            self.json_files.extend(extra_jsons)
        
        self.json_files = list(set(self.json_files))  # deduplicate
        self.tdenv.DEBUG1("elite game json files: {!r}", self.json_files)

        # self.loaded_time is set on completion of self.refresh()
        
        with bench(self.tdenv, "EliteGame refresh"):
            self.refresh()
    
    def list_journals(self) -> list[Path]:
        """ Returns a list of journals with the most recent first, if any are present in the journal path. """
        journals: dict[Path, str] = {}
        candidates: list[Path] = list(self.path.glob("Journal.*.log"))  # otherwise it's a generator

        self.tdenv.DEBUG3("available: {}", candidates)
        for journal_file in candidates:
            # It's expected to have YYYY'-'MM'-'DD'T'HHMMSS'.'LOGN'.log',
            # which forms a nice simple collation order number/string
            index = re.match(r"Journal\.(\d{4})-(\d{2})-(\d{2})T(\d{6})\.(\d+)\.log$", journal_file.name)
            if not index:
                self.tdenv.DEBUG2("ignoring {}, doesn't match our guard pattern", journal_file)
                continue  # ignore, no match
            
            self.tdenv.DEBUG3("considering {}", journal_file)
            journals[journal_file] = "".join(index.groups())
        
        return sorted(journals.keys(), key=lambda file: journals[file], reverse=True)
    
    def refresh(self) -> None:
        """
        Refresh game state by re-reading journal and status files.
        
        Call this method to update state after game events occur. Useful
        when polling for changes in a long-running application.
        """
        self.tdenv.DEBUG1("Refreshing game state")
        self.status = None
        with bench(self.tdenv, "_load_journals"):
            self._load_journals()
        with bench(self.tdenv, "_load_json_files"):
            self._load_json_files()
        self.loaded_time = datetime.datetime.now()
    
    def _load_journals(self) -> None:
        """
            @internal _load_journals will apply self.journal_load and
            attempt to have the appropriate journal files loaded.
            
            If self.journal_load is DO_NOT_LOAD, no files are loaded and existing
            data is cleared. If MOST_RECENT, only the most recent journal is
            loaded. If ALL_JOURNALS, all available journals are loaded. If an
            integer value greater than 1 is provided, upto that many are loaded -
            most recent first.
        """
        # Use the journal_load to determine how many journal files we're
        # supposed to read.
        match self.journal_load:
            case JournalLoad.ALL_JOURNALS:
                max_journals = -1

            case JournalLoad.DO_NOT_LOAD | int() as n if n <= 0:
                self.journals_loaded.clear()
                self.journal.clear()
                self.events.clear()
                self.tdenv.DEBUG1("JournalLoad.DO_NOT_LOAD; cleared journals")
                return

            case JournalLoad.MOST_RECENT | int() as n if n == 1:
                max_journals = 1

            case int() as n:
                max_journals = n

            case _:
                raise GameException(f"Invalid journal load value: {self.journal_load}")
        
        # Sanity check: zero means zero, ie don't load anything.
        if max_journals == 0:
            raise RuntimeError("Internal error: zero max_journals escaped match")
        
        # Fetch the full list of files
        files = self.list_journals()
        if not files:
            raise JournalException(f"no journal logs found in {self.path}")

        # Positive count denotes an upper cound
        if max_journals > 0:
            files = files[:max_journals]
        
        # Now we need them in the reverse order so that newer values overwrite older.
        files.reverse()
        
        self.tdenv.DEBUG1("loading {} journal files: {}", len(files), files)
        with bench(self.tdenv, "_load_journal_files"):
            try:
                self._load_journal_files(files)
            except FileNotFoundError as e:
                raise JournalException(str(e)) from None

        if not self.journals_loaded:
            raise JournalException(f"no valid journal logs found in {self.path}")
    
    def _add_event(self, event: dict[str, Any] | None) -> None:
        """ @internal helper for registering an event into our fields """
        event_name = (event or {}).get("event", None)
        if not event_name:
            self.tdenv.DEBUG2("ignoring non-event: {}", event)
            return
        
        self.journal.append(event)
        try:
            self.history[event_name].append(event)
        except KeyError:
            self.history[event_name] = [event]
        self.events[event_name] = event
    
    def _load_journal_files(self, files: list[Path], *, loader: Any = orjson.loads) -> None:
        """
            @internal _load_journal_files loads events from the provided list of
            journal files into self.journals_loaded, self.journal, self.history,
            and self.events.
            @note silo'd for testability.
        """
        for journal_file in files:
            self.tdenv.DEBUG1("loading journal file: {}", journal_file)
            with journal_file.open("r", encoding="utf-8") as line_stream:
                lines = iter(line_stream)
                try:
                    header = loader(next(lines))
                except (StopIteration, orjson.JSONDecodeError):
                    raise JournalException("Unable to read journals: Please wait for Elite Dangerous to finish loading") from None
                
                try:
                    # Use an iterator so we can siphon off the first line as a header
                    if not self._read_journal_header(journal_file, header):
                        continue
                    self.tdenv.DEBUG0(
                        "{}: elite {} build {} lang {} odyssey {}",
                        journal_file,
                        header.get("gameversion", "unknown"),
                        header.get("build", "unknown"),
                        header.get("language", "unknown"),
                        header.get("Odyssey", "unknown")
                    )
                    
                    for line in lines:
                        data = loader(line)
                        self._add_event(data)
                    
                    self.journals_loaded.append(journal_file)
                
                except orjson.JSONDecodeError as e:
                    self.tdenv.WARN("failed to parse journal file {}: {}", journal_file, e)
    
    def _read_journal_header(self, journal_file: Path, header: dict) -> bool:
        """
            @internal _read_journal_header reads and validates the header line
            of a journal file. Returns True if successful, False to skip the file.
            @note silo'd for testability.
        """
        if not isinstance(header, dict) or header.get("event", "") != "Fileheader":
            self.tdenv.DEBUG0("ignoring invalid journal file {}, first line is not Fileheader: {}", journal_file, header)
            return False
        
        event: str = header["event"]
        if event not in self.history:
            self.history[event] = []
        self._add_event(header)
        
        return True
    
    def _load_json_files(self) -> None:
        if not self.json_files:
            self.json_data.clear()
            self.jsons_loaded.clear()
            self.tdenv.DEBUG1("no json files to load; cleared json data")
            return
        
        for json_file_enum in set(self.json_files):
            json_file_path = self.path / json_file_enum.value
            self.tdenv.DEBUG1("loading json file: {}", json_file_path)

            try:
                with json_file_path.open("r", encoding="utf-8") as json_stream:
                    data = orjson.loads(json_stream.read())
                    self.json_data[json_file_enum] = data
                    self.jsons_loaded += [json_file_enum]
                    self._add_event(data)
            except FileNotFoundError as e:
                raise MissingJsonError(str(e)) from None
            except orjson.JSONDecodeError as e:
                raise JsonException(f"unable to parse save-game json file {json_file_path}: {e}") from None
    
    def get_status(self) -> Status:
        """
            get_status creates or returns a cached summary of the
            already-loaded json status.
        """
        if self.status:
            return self.status
        
        # Start a new one
        status = Status()

        # Collect journal events.
        for event in self.journal:
            match event.get("event"):
                case "Location" | "FSJump" | "CarrierJump":
                    status.star_system = event.get("StarSystem") or status.star_system  # incase it's empty
                    status.system_address = event.get("SystemAddress", status.system_address)
                    status.docked = event.get("Docked", status.docked)
                    status.landed = event.get("Landed", status.landed)
                    if status.docked:
                        status.station_name = event.get("StationName", status.station_name)
                        status.station_type = event.get("StationType", status.station_type)

                case "Cargo":
                    if event.get("Vessel") == "Ship":
                        status.cargo_load = event.get("Count")
                
                case "Docked":
                    status.docked = True
                    status.landed = False
                    status.station_name = event.get("StationName", status.station_name)
                    status.station_type = event.get("StationType", status.station_type)
                    status.star_system = event.get("StarSystem", status.star_system)
                    status.system_address = event.get("SystemAddress", status.system_address)

                case "Loadout":
                    status.cargo_space = event.get("CargoCapacity")
                
                case "Undocked":
                    status.docked = False
                    status.station_name = None
                    status.station_type = None
                
                case "Touchdown" | "Landed":
                    status.docked = False
                    status.landed = True
                
                case "Liftoff":
                    status.docked = status.landed = False
                
                case "Commander":
                    status.commander = event.get("Name")
                    status.fid = event.get("FID")
                
                case "LoadGame":
                    status.commander = event.get("Commander")
                    status.fid = event.get("FID")
                
                case "Status":
                    status.credits = event.get("Balance")
                    status.destination = event.get("Destination")
                
                case _:
                    continue  # TERMINATE
            
            if not status.docked and not status.landed:
                status.station_name = status.station_type = None
            
            status.timestamp = event.get("timestamp", status.timestamp)
        
        self.status = status
        
        return status


def journal_path(tdenv: TradeEnv, path: str | Path | None) -> Path:
    """ Attempts to locate the elite savegame folder. """
    candidate: Path | str | None = None
    source: str | None = None
    
    if path:
        # the caller is providing an explicit path, they win.
        source, candidate = "path", path
    elif (env_path := getattr(tdenv, "journal_path", None)):
        # configured on the context object, it wins
        source, candidate = "tdenv.journal_path", env_path
    elif (env_path := os.getenv(ELITE_JOURNAL_ENVVAR, None)):
        # environment variable fallback
        source, candidate = ELITE_JOURNAL_ENVVAR, env_path
    elif WINDOWS_JOURNAL_DIR.exists():
        source, candidate = "windows", WINDOWS_JOURNAL_DIR
    elif MACOS_JOURNAL_DIR.exists():
        source, candidate = "macos", MACOS_JOURNAL_DIR
    
    if not candidate:
        raise GameFolderError("Could not determine the default path for this machine")
    
    tdenv.DEBUG0("journal_path: {}: {}", source, path)
    
    path = Path(candidate)
    if not path.exists():
        raise GameFolderError(f"'{candidate}' does not exist")
    
    return path


def require_game_data(
    game: EliteGame,
    *,
    cargo: bool = False,        # Short for cargo_space + cargo_load
    cargo_space: bool = False,  # Size of cargo hold
    cargo_load: bool = False,   # Current units of cargo
    commander: bool = False,
    credits: bool = False,
    location: bool = False,
    navroute: bool = False,
    status_fields: Iterable[str] | None = None,  # list of fields you need status to have.
) -> None:
    """ Checks the game object has the data it needs to provide
        specific status fields, and if not raises an exception
        with user guidance. """
    # Fields that we'll want to check aren't "None"
    none_checks: set[str] = set()  # not to be confused with nun_chucks

    #### EVENTS

    # We require that one item from each subset is present,
    # that is: all(any(subset) for subset in events)
    events: set[str | Iterable[str]] = set()

    if cargo or cargo_space:
        events.add(("Loadout",))    # The ship's load out, i.e its capacity
        none_checks.add("cargo_space")
    if cargo or cargo_load:
        events.add(("Cargo",))      # The ship's cargo content, i.e cargo hold use
        none_checks.add("cargo_load")

    if commander:
        events.add(("Commander", "LoadGame"))  # either will do
        none_checks.add("commander")

    if credits:
        events.add(("Status",))
        none_checks.add("credits")

    if location:
        events.add((
            "CarrierJump",
            "Docked",
            "FSJump",
            "Location",
        ))

    missing = set()
    for subset in events:
        has_any = any(event in game.events for event in subset)
        if has_any:
            continue
        for event in subset:
            missing.add(event)

    #### JSON loads/events
    if navroute:
        if not game.json_data.get(JsonFiles.NAVROUTE, {}):
            missing.add("NavRoute data")

    #### None checks

    # Now we'll check that there are meaningful values on fields the caller wants to access.
    status: Status = game.get_status()

    check_attrs = none_checks | set(status_fields or ())
    for field in check_attrs:
        try:
            if getattr(status, field) is None:
                missing.add(f"'{field}' value")
        except AttributeError:
            raise RuntimeError(f"Internal Error: Checking for unrecognized game-data field '{field}'")

    if missing:
        missing_things = sorted(missing)
        raise GameDataException(
            "Your current game is missing information required to "
            f"serve this request: {', '.join(missing_things)}"
        )



from ..tradeexcept import TradeException


######################################################################
# Exceptions

class UsageError(TradeException):
    def __init__(self, title, usage):
        self.title, self.usage = title, usage
    
    def __str__(self):
        return f"{self.title}\n\n{self.usage}"


class CommandLineError(TradeException):
    """
        Raised when you provide invalid input on the command line.
        Attributes:
            errorstr       What to tell the user.
    """
    def __init__(self, errorStr, usage=None):
        self.errorStr, self.usage = errorStr, usage
    
    def __str__(self):
        if self.usage:
            return f"ERROR: {self.errorStr}\n\n{self.usage}"
        return f"ERROR: {self.errorStr}"


class NoDataError(TradeException):
    """
        Raised when a request is made for which no data can be found.
        Attributes:
            errorStr        Describe the problem to the user.
    """
    def __init__(self, errorStr):
        self.errorStr = errorStr
    
    def __str__(self):
        return f"""Error: {self.errorStr}
Possible causes:
- No profitable runs or routes meet your criteria,
- Missing Systems or Stations along the route (see "local -vv"),
- Missing price data (see "market -vv" or "update -h"),

If you are not sure where to get data from, consider using a crowd-sourcing
project such as EDDBlink (https://github.com/eyeonus/EDDBlink).

For more help, see the TradeDangerous Wiki:
    https://github.com/eyeonus/Trade-Dangerous/wiki
"""

class GameDataError(TradeException):
    """
        Raised when imported or journal data is internally inconsistent
        or clearly invalid (e.g. cargo load exceeds cargo capacity).

        Attributes:
            errorStr    A short description of the inconsistency.
    """
    def __init__(self, errorStr):
        self.errorStr = errorStr

    def __str__(self):
        return f"""Error: {self.errorStr}
Possible causes:
- The journal data was incomplete or out of sync when read,
- You recently changed ships or game state and journals are still updating,
- The wrong journals or save files are being read (check your journal path),
- Or the data on disk has been corrupted.

Try again after the game has fully updated, or re-run the command with
--full-load to ignore current cargo occupancy if appropriate.

For more help, see the TradeDangerous Wiki:
    https://github.com/eyeonus/Trade-Dangerous/wiki
"""

class PadSizeError(CommandLineError):
    """ Raised when an invalid pad-size option is given. """
    def __init__(self, value):
        super().__init__(
            f"Invalid --pad-size '{value}': Use a combination of one or more "
            "from 'S' for Small, 'M' for Medium, 'L' for Large or "
            "'?' for unknown, e.g. 'SML?' matches any pad size while "
            "'M?' matches medium or unknown or 'L' matches only large."
        )


class PlanetaryError(CommandLineError):
    """ Raised when an invalid planetary option is given. """
    def __init__(self, value):
        super().__init__(
            f"Invalid --planetary '{value}': Use a combination of one or more "
            "from 'Y' for Yes, 'N' for No or '?' for unknown, "
            "e.g. 'YN?' matches any station while 'Y?' matches "
            "yes or unknown, or 'N' matches only non-planetary stations."
        )


class FleetCarrierError(CommandLineError):
    """ Raised when an invalid fleet-carrier option is given. """
    def __init__(self, value):
        super().__init__(
            f"Invalid --fleet-carrier '{value}': Use a combination of one or more "
            "from 'Y' for Yes, 'N' for No or '?' for unknown, "
            "e.g. 'YN?' matches any station while 'Y?' matches "
            "yes or unknown, or 'N' matches only non-fleet-carrier stations."
        )

class OdysseyError(CommandLineError):
    """ Raised when an invalid odyssey option is given. """
    def __init__(self, value):
        super().__init__(
            f"Invalid --odyssey '{value}': Use a combination of one or more "
            "from 'Y' for Yes, 'N' for No or '?' for unknown, "
            "e.g. 'YN?' matches any station while 'Y?' matches "
            "yes or unknown, or 'N' matches only non-odyssey stations."
        )



# tradedangerous/commands/station_cmd.py — DEPRECATED (no-op)
# This command no longer edits station records directly.
#
# How to maintain station data going forward:
#   • Import authoritative data via:   trade import -P spansh  |  trade import -P eddblink
#   • Solo/offline players (manual capture via EDMC plugin):
#       TradeDangerous DB-Update for EDMC → https://github.com/bgol/UpdateTD
#
# To inspect a station’s data, use read-only commands like:
#   trade station <name>         ← (legacy pattern; now deprecated)
#   trade local --near "<system>"   (discoverability)
#   trade market --origin "<station>" (market view)

from __future__ import annotations
from .parsing import ParseArgument

# ---- Command metadata ----
help = "DEPRECATED: no longer used. See deprecation banner when run."
name = "station"
epilog = None
acceptUnknown = True

# No DB access is needed for this no-op command.
wantsTradeDB = False
usesTradeData = False

# Accept ANY number of positional args and ignore them (prevents parser errors).
arguments = (
    ParseArgument(
        'args',
        help="(deprecated) ignored",
        nargs='*',
        type=str,
    ),
)
# No switches; unknown switches will still be rejected by the global parser.
switches = (
)


def _banner() -> str:
    return (
        "\n"
        "=== DEPRECATION NOTICE: station ====================================\n"
        "This command no longer edits station records and does not modify the DB.\n"
        "• Import station data via:    trade import -P eddblink | -P spansh\n"
        "• Solo/offline capture via EDMC plugin:\n"
        "    TradeDangerous DB-Update → https://github.com/bgol/UpdateTD\n"
        "=====================================================================\n"
    )


def run(results, cmdenv, tdb=None):
    """
    No-op implementation: print banner and exit immediately.
    All arguments/switches are ignored by design.
    """
    banner = _banner()
    try:
        cmdenv.NOTE("{}", banner)
    except Exception:
        print(banner)
    return False  # Nothing to render


def render(results, cmdenv, tdb=None):
    # No output beyond the banner emitted in run().
    return



#! /usr/bin/env python

"""
based on edsc.py
uses EDSM - https://www.edsm.net/api
"""

import json
import requests


def edsm_log(apiCall, url, params, jsonData=None, error=None):
    """
    Logs an EDSM query and it's response to tmp/edsm.log
    """
    try:
        with open("tmp/edsm.log", "a", encoding="utf-8") as fh:
            print('-'*70, file=fh)
            print("API:", apiCall, file=fh)
            print("REQ:", str(params), file=fh)
            print("URL:", url, file=fh)
            if jsonData:
                print("REP:", json.dumps(jsonData, indent=1), file=fh)
            if error:
                print("ERR:", error)
            print(file=fh)
    except FileNotFoundError:
        pass


class EDSMQueryBase:
    """
    Base class for creating an EDSM Query class, do not use directly.
    
    Derived class must declare "apiCall" which is appended to baseURL
    to form the query URL.
    """
    
    baseURL = "https://www.edsm.net/api-v1/"
    
    def __init__(self, log=False, known=1, coords=1, **kwargs):
        self.log = log
        self.url = self.baseURL + self.apiCall
        self.params = {
            'known': known,
            'coords': coords,
        }
        for k, v in kwargs.items():
            self.params[k] = v
    
    def fetch(self):
        res = requests.get(self.url, self.params)
        self.status = res.status_code
        
        try:
            data = res.json()
        except:
            data = None
            pass
        
        if self.log:
            edsm_log(self.apiCall, res.url, self.params, data)
        
        return data


class StarQuerySingle(EDSMQueryBase):
    """
    Query EDSM System.
    
    provide:
        systemName
    """
    apiCall = "system"

class StarQueryMulti(EDSMQueryBase):
    """
    Query EDSM Systems.
    
    provide:
        systemName
    """
    apiCall = "systems"

class StarQuerySphere(EDSMQueryBase):
    """
    Query EDSM Systems.
    provide:
        systemName or center-coords as x, y and z
        radius N.NN
    """
    apiCall = "sphere-systems"

class DistanceQuery(EDSMQueryBase):
    """
    Request distances from EDSM.
    """
    apiCall = "distances"


if __name__ == "__main__":
    print("Requesting Azrael, coords-known")
    edsq = StarQuerySingle(log=True, systemName="Azrael")
    data = edsq.fetch()
    print("{:<30s} {:11f} {:11f} {:11f}".format(
        data['name'].upper(),
        data['coords']['x'],
        data['coords']['y'],
        data['coords']['z'],
    ))
    
    print("Requesting 10ly sphere of Sol, coords-known")
    edsq = StarQuerySphere(systemName="Sol", radius=10)
    data = edsq.fetch()
    for sysinfo in data:
        print("{:<30s} {:11f} {:11f} {:11f}".format(
            sysinfo['name'].upper(),
            sysinfo['coords']['x'],
            sysinfo['coords']['y'],
            sysinfo['coords']['z'],
        ))



#!/usr/bin/env python3.6
# Deprecated
# Website no longer exists

"""
Small tool to submit new star data to EDStarCoordinator.

Use:
    submit-distances.py "system name"
"""

#
# NOTE: This tool is very hacky. If someone wants to clean it up,
# and submit a diff, that'd be greatly appreciated!
#

import argparse
import os
import random
import re
import requests
import sys
import tradedb
import tradeenv

from tradedangerous.misc.edsc import StarSubmission, StarSubmissionResult, SubmissionError
from tradedangerous.misc.clipboard import SystemNameClip


standardStars = [
    "SOL",
    "NEW YEMBO",
    "VESUVIT",
    "HIP 79884",
    "ASELLUS AUSTRALIS",
]


sys.stderr.write("*** WARNING: submit-distances.py is deprecated; if you rely on it, please post a github issue\n")


############################################################################


class UsageError(Exception):
    def __init__(self, argv, error):
        self.argv, self.error = argv, error
    
    def __str__(self):
        return error + "\n" + argv.format_usage()

def parse_arguments():
    parser = argparse.ArgumentParser(
        description='Submit star distances to the EDSC project.',
    )
    parser.add_argument(
        'origin',
        help='System to submit distances for.',
        type=str,
    )
    parser.add_argument(
        '--pick',
        help='Randomly select N existing systems from gamma or ealier',
        required=False,
        type=int,
    )
    parser.add_argument(
        '--cmdr',
        required=False,
        help='Specify your commander name.',
        type=str,
        default=os.environ.get('CMDR', None)
    )
    parser.add_argument(
        '--test',
        help='Use the EDSC Test Database.',
        required=False,
        action='store_true',
    )
    parser.add_argument(
        '--detail', '-v',
        help='Output additional detail.',
        action='count',
        default=0,
    )
    parser.add_argument(
        '--debug', '-w',
        help='Enable debugging output,',
        action='count',
        default=0,
    )
    parser.add_argument(
        '--no-update',
        help='Disallow distances from an existing system.',
        action='store_false',
        default=True,
        dest='allowUpdate',
    )
    parser.add_argument(
        '--extra-file',
        help='File to read/write extra stars to.',
        type=str,
        default='data/extra-stars.txt',
        dest='extraFile',
    )
    parser.add_argument(
        'destinations',
        help='System or systems to measure distance to.',
        default=[],
        nargs='*',
    )
    
    argv = parser.parse_args(sys.argv[1:])
    argv.origin = argv.origin.upper()
    if argv.origin.startswith('@'):
        argv.origin = argv.origin[1:]
    
    if not argv.cmdr:
        raise UsageError(argv, "No commander name specified")
    
    return argv

def get_system(argv, tdb):
    system = tdb.systemByName.get(argv.origin.upper(), None)
    if not system:
        return argv.origin, None
    
    if not argv.allowUpdate:
        raise UsageError(
            argv,
            "System '{}' already exists.\n"
            .format(systemName)
        )
    
    if argv.detail:
        print("EXISTING SYSTEM:", argv.origin)
    
    return argv.origin, system

def pick_destinations(argv, tdb):
    numSystems = len(tdb.systemByName)
    if numSystems < 1:
        raise UsageError(
            argv,
            "Can't --pick random systems: "
            "Your TD database doesn't contain any systems."
        )
    num = min(argv.pick, numSystems)
    systems = tdb.systemByName
    try:
        gamma = tdb.lookupAdded("Gamma")
    except KeyError:
        gamma = 20
    destinations = random.sample([
        sysName for sysName, system in systems.items()
        if system.addedID <= gamma and \
            not sysName in standardStars
    ], num)
    
    return destinations

def get_outliers(argv):
    outliers = set()
    try:
        with open(argv.extraFile, "r", encoding="utf-8") as input:
            for line in input:
                name = line.partition('#')[0].strip().upper()
                if name and name != argv.origin:
                    outliers.add(name)
    except FileNotFoundError:
        pass
    outliers = list(outliers)
    random.shuffle(outliers)
    return outliers

def get_distances(argv, clip, stars):
    distances = []
    for star in stars:
        starNo = len(distances) + 1
        # Check it's not already in the list
        star = star.upper()
        if star in distances:
            continue
        
        clip.copy_text(star)
        
        if len(stars) > 1:
            prefix = "{:>2}/{:2}: ".format(starNo, len(stars))
        else:
            prefix = ""
        dist = input(prefix + "Distance to {}: ".format(star))
        if dist == 'q':
            return distances, 'q'
        
        if dist:
            distances.append({
                'name': star.upper(),
                'dist': float(dist),
            })
    
    return distances, 'end'

def check_system(argv, tdb, tdbSys, name):
    system = tdb.systemByName.get(name.upper(), None)
    if system and not tdbSys:
        print("KNOWN SYSTEM")
        return
    if system:
        print("KNOWN SYSTEM: {:.2f} ly".format(
            tdbSys.distanceTo(system)
        ))

def add_extra_stars(argv, extraStars):
    if not extraStars:
        return
    if argv.detail:
        print("Saving {} to {}".format(
            str(extraStars), argv.extraFile,
        ))
    try:
        with open(argv.extraFile, "a", encoding="utf-8") as output:
            for star in extraStars:
                print(star, file=output)
    except FileNotFoundError:
        pass

def submit_distances(argv, tdb, distances):
    system = argv.origin
    cmdr = argv.cmdr
    mode = "TEST" if argv.test else "Live"
    
    print()
    print("System:", system)
    print("Database:", mode)
    print("Distances:")
    for ref in distances:
        print("  {}: {:.02f} ly".format(
            ref['name'], ref['dist']
        ))
    print()
    
    ok = input("Does this look correct (y/n)? ")
    if ok != 'y':
        print("Stopped")
        return
    
    print()
    print("Submitting {} {}".format(mode, system))
    
    sub = StarSubmission(
        star=system,
        commander=cmdr,
        refs=distances,
        test=argv.test,
    )
    resp = sub.submit()
    
    result = StarSubmissionResult(star=system, response=resp)
    print(str(result))
    if result.valid:
        # Check for systems we can add
        trilats = set()
        for sysName in result.systems.keys():
            code, coord = result.systems[sysName]
            sysName = sysName.upper()
            # Does it have a distance
            if isinstance(coord, (list, tuple)):
                x, y, z = coord
                system = tdb.systemByName.get(sysName, None)
                if system:
                    tdb.updateLocalSystem(system, sysName, x, y, z)
                else:
                    tdb.addLocalSystem(sysName, x, y, z)
        if result.recheck:
            return list(result.recheck.keys())
    return None

def do_rechecks(argv, clip, rechecks):
    print("\aSome systems need their distances rechecked:")
    
    distances, term = get_distances(argv, clip, rechecks)
    return distances

def send_and_check_distances(argv, tdb, clip, distances):
    if not distances:
        if argv.detail:
            print("No distances, no submission.")
        return False
    
    while distances:
        rechecks = submit_distances(argv, tdb, distances)
        if not rechecks:
            break
        
        distances = do_rechecks(argv, clip, rechecks)
    
    return True

def process_destinations(argv, tdb):
    clip = SystemNameClip()
    
    print("Distances from {}:".format(argv.origin))
    distances, _ = get_distances(argv, clip, argv.destinations)
    send_and_check_distances(argv, tdb, clip, distances)

############################################################################

def main():
    argv = parse_arguments()
    
    tdenv = tradeenv.TradeEnv(properties=argv)
    tdb = tradedb.TradeDB(tdenv)
    
    system, tdbSys = get_system(argv, tdb)
    
    if argv.pick:
        argv.destinations.extend(pick_destinations(argv, tdb))
    
    if argv.destinations:
        process_destinations(argv, tdb)
        return
    
    print("Add EDSC Star Distances for \"{}\"".format(system))
    print()
    print("You will now be prompted for distances to various stars.")
    print()
    print(
        "At each prompt, the star name will be copied into your paste buffer. "
        "You should alt-tab into the game and paste the name into the Galaxy "
        "Map's search box. Then alt-tab back and enter the distance value."
    )
    print()
    print(
        "At each prompt enter a ly distance (e.g. 123.45), q to stop, "
        "or leave the line empty if you don't want to enter data for "
        "this star."
    )
    print(
        "5 distances are required for EDSC to make a first guess at a "
        "star's location. You can submit more to increase the accuracy "
        "but the only time you need to submit more than 10 is when you "
        "are trying to submit corrections."
    )
    print()
    
    clip = SystemNameClip()
    
    print()
    print("""
===================================================
STANDARD STARS: (q to skip to the next section)
  
  These are stars with well-known positions.
===================================================
""")
    distances, term = get_distances(argv, clip, standardStars)
    if distances:
        send_and_check_distances(argv, tdb, clip, distances)
    
    outliers = get_outliers(argv)
    if outliers:
        print("""
===================================================
EXTRA STARS: (q to skip to the next section)
  
  Stars from {}.
===================================================
""".format(argv.extraFile))
        distances, term = get_distances(argv, clip, outliers)
        if distances:
            send_and_check_distances(argv, tdb, clip, distances)
    
    print("""
===================================================
CHOOSE YOUR OWN: (q to stop)
  
  Specify additional stars.
  
  Prefix names with a '+' to add them to
  {}.
===================================================
""".format(argv.extraFile))
    distances = []
    newOutliers = []
    while True:
        star = input("Enter star name: ")
        star = star.strip().upper()
        if not star or star == 'Q':
            break
        # Remove surrounding quotes
        save = False
        if star.startswith('+'):
            save = True
            star = star[1:].strip()
        star = re.sub(r'\s+', ' ', star)
        star = re.sub(r"''+", "'", star)
        star = re.sub(r'^("|\')+\s*(.*)\s*\1+', r'\2', star)
        if star.find('"') >= 0:
            print("Invalid star name")
            continue
        if star.startswith('+'):
            save = True
            star = star[1:].strip()
        for ref in distances:
            if ref['name'] == star:
                print("'{}' is already listed.")
                continue
        check_system(argv, tdb, tdbSys, star)
        extras, term = get_distances(argv, clip, [star])
        if term != 'q' and len(extras) > 0:
            distances.extend(extras)
            if save and star not in outliers and star not in newOutliers:
                newOutliers.append(star)
    
    if send_and_check_distances(argv, tdb, clip, distances):
        add_extra_stars(argv, newOutliers)

if __name__ == "__main__":
    try:
        main()
    except (SubmissionError, UsageError) as e:
        print(str(e))



# Deprecated
import tkinter as tk
import tkinter.messagebox as mbox
import sqlite3
import re

"""
This is a crude attempt at a GUI for updating trade prices.
It needs a lot of love to be less... Shitty.

TODO:
. Add a way to add a new line,
. Add a way to reorder lines,
. Add a Save button,
. Add pre-exit validation (and don't exit if there's a problem),

Thanks to Bryan Okler for solving the focus tracking issue:
http://stackoverflow.com/questions/27055936/python3-tk-scrollbar-and-focus

"""

updateUiHelp = (
"ABOUT THE EDITOR:\n"
"The columns shown here represent the columns in the "
"in-game Commodites Market UI and are in the same order.\n"
"\n"
"WHAT TO ENTER:\n"
"Leave items blank if they are NOT listed in the game UI.\n"
"\n"
"'Demand' is disabled unless you use '--use-demand'.\n"
"\n"
"'Supply' should be:\n"
"  '-' or '0' if no units are available,\n"
"  empty or '?' if the level is unknown,\n"
"  or the amount of supply followed by L, M or H.\n"
"E.g.\n"
"  1L, 50M, 3000H.\n"
"\n"
"SAVING:\n"
"To save your data, click one of the window's close buttons.\n"
"\n"
"NAVIGATION:\n"
"- Use Tab, Shift-Tab, Up/Down Arrow and Enter to navigate.\n"
)

class Item:
    """ Describe a listed, tradeable item """
    
    def __init__(self, ID, catID, name, displayNo):
        self.ID, self.catID, self.name = ID, catID, name
        self.displayNo = displayNo


class ScrollingCanvas(tk.Frame):
    """
        Tk.Canvas with scrollbar.
    """
    
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        
        self.canvas = canvas = tk.Canvas(self, borderwidth=0)
        canvas.grid(row=0, column=0)
        canvas.grid_rowconfigure(0, weight=1)
        canvas.grid_columnconfigure(0, weight=1)
        
        vsb = tk.Scrollbar(parent,
                    orient=tk.VERTICAL,
                    command=canvas.yview)
        vsb.grid(row=0, column=1)
        vsb.pack(side="right", fill=tk.BOTH, expand=False)
        
        canvas.configure(yscrollcommand=vsb.set)
        canvas.configure(yscrollincrement=4)
        
        self.interior = interior = tk.Frame(canvas)
        canvas.create_window(0, 0, window=interior, anchor="nw", tags="interior")
        canvas.bind("<Configure>", self.onConfigure)
        interior.rowconfigure(0, weight=1)
        interior.columnconfigure(0, weight=1)
        interior.columnconfigure(1, weight=1)
        interior.columnconfigure(2, weight=1)
        interior.columnconfigure(3, weight=1)
        interior.columnconfigure(4, weight=1)
        
        canvas.grid(row=0, column=0)
        canvas.pack(side="left", fill=tk.BOTH, expand=True)
        
        self.bind_all("<MouseWheel>", self.onMouseWheel)
        canvas.bind_all("<FocusIn>", self.scrollIntoView)
        
        self.pack(side="left", fill=tk.BOTH, expand=True)


    def onConfigure(self, event):
        """ Handle resizing """
        
        self.canvas.configure(scrollregion=self.interior.bbox("all"))


    def onMouseWheel(self, event):
        """ Translate mouse wheel inputs to scroll bar motions """
        
        self.canvas.yview_scroll(int(-1 * (event.delta/4)), "units")


    def scrollIntoView(self, event):
        """ Scroll so that a given widget is in focus """
        
        canvas = self.canvas
        widget_top = event.widget.winfo_y()
        widget_bottom = widget_top + event.widget.winfo_height()
        canvas_top = canvas.canvasy(0)
        canvas_bottom = canvas_top + canvas.winfo_height()
        
        if widget_bottom >= canvas_bottom:
            delta = int(canvas_bottom - widget_bottom)
            canvas.yview_scroll(-delta, "units")
        elif widget_top < canvas_top:
            delta = int(widget_top - canvas_top)
            canvas.yview_scroll(delta, "units")


class UpdateGUI(ScrollingCanvas):
    """
        Implements a tk canvas which displays an editor
        for TradeDangerous Price Updates
    """
    
    def __init__(self, parent, tdb, tdenv):
        super().__init__(parent)
        
        width = tdenv.width or 640
        height = tdenv.height or 640
        sticky = 1 if tdenv.alwaysOnTop else 0
        
        self.tdb = tdb
        self.tdenv = tdenv
        
        self.rowNo = 0
        self.colNo = 0
        self.items = {}
        self.categories = []
        self.itemList = []
        self.itemDisplays = []
        self.results = None
        self.headings = []
        
        self.createWidgets()
        
        self.focusOn(0, 0)
        
        parent.geometry("{}x{}{:+n}{:+n}".format(
                    width+16, height,
                    tdenv.windowX,
                    tdenv.windowY,
                ))
        
        # Allow the window to be always-on-top
        parent.wm_attributes("-topmost", sticky)


    def selectWidget(self, widget, newValue=None):
        if newValue is not None:
            widget.val.set(newValue)
        widget.focus_set()
        widget.selection_range(0, tk.END)


    def focusOn(self, displayNo, pos):
        """ Set focus to a widget and select the text in it """
        
        row = self.itemDisplays[displayNo]
        widget = row[pos][0]
        self.selectWidget(widget)


    def setValue(self, row, pos, value):
        row[pos][1].set(value)
        return value


    def getValue(self, row, pos):
        return row[pos][1].get()


    def validateRow(self, row):
        paying = row[0][1].get()
        if not paying or paying == "0":
            paying = ""
            for col in row:
                col[1].set("")
            row[1][0].configure(state=tk.DISABLED)
            row[2][0].configure(state=tk.DISABLED)
        else:
            row[1][0].configure(state=tk.NORMAL)
            if self.tdenv.useDemand:
                row[2][0].configure(state=tk.NORMAL)
            else:
                row[2][0].configure(state=tk.DISABLED)
                row[2][1].set("")
        
        asking = row[1][1].get()
        if asking == "0":
            row[1][1].set("")
            asking = ""
        if not asking:
            row[3][0].configure(state=tk.DISABLED)
            row[3][1].set("")
        else:
            row[3][0].configure(state=tk.NORMAL)
            if not row[3][1].get():
                row[3][1].set("?")


    def checkValueAgainstStats(self, widget, stats):
        widget.configure(bg = 'white')
        minCr, avgCr, maxCr = stats
        if not avgCr:
            return True
        
        cr = int(widget.val.get())
        
        if cr < int(minCr / 1.01) and cr < int(avgCr * 0.7):
            widget.bell()
            ok = mbox.askokcancel(
                    "Very low price",
                    "The price you entered is very low.\n"
                    "\n"
                    "Your input was: {}\n"
                    "Previous low..: {}\n"
                    "\n"
                    "Is it correct?".format(
                        cr, minCr,
            ))
            if not ok:
                self.selectWidget(widget, "")
                return False
            widget.configure(bg = '#ff8080')
        
        if cr >= (maxCr * 1.01) and int(cr * 0.7) > avgCr:
            widget.bell()
            ok = mbox.askokcancel(
                    "Very high price",
                    "The price you entered is very high.\n"
                    "\n"
                    "Your input was..: {}\n"
                    "Previous highest: {}\n"
                    "\n"
                    "Is it correct?".format(
                        cr, maxCr,
            ))
            if not ok:
                self.selectWidget(widget, "")
                return False
            widget.configure(bg = '#8080ff')
        
        return True


    def validate(self, widget):
        """ For checking the contents of a widget. """
        item, row, pos = widget.item, widget.row, widget.pos
        
        value = widget.val.get()
        re.sub(" +", "", value)
        re.sub(",", "", value)
        widget.val.set(value)
        
        self.validateRow(row)
        
        if pos == 0:
            if not value or value == "0":
                widget.val.set("")
                return True
            
            if not re.match(r'^\d+$', value):
                mbox.showerror(
                        "Invalid Paying price",
                        "Price must be an numeric value (e.g. 1234)"
                        )
                return False
            
            # is it lower than the value?
            demandStats = self.demandStats[item.ID]
            if not self.checkValueAgainstStats(widget, demandStats):
                return False
        
        if pos <= 1:
            if not value or value == "0":
                widget.val.set("")
                return True
            
            if not re.match(r'^\d+$', value):
                mbox.showerror(
                        "Invalid Paying price",
                        "Price must be an numeric value (e.g. 1234)"
                        )
                return False
        
        if pos == 1:
            # Don't allow crazy difference between prices
            paying = int(row[0][1].get())
            asking = int(row[1][1].get())
            
            if asking < paying:
                widget.bell()
                mbox.showerror(
                        "I DOUBT THAT!",
                        "Stations never pay more for an item than "
                        "they sell it for.",
                        )
                return False
            
            supplyStats = self.supplyStats[item.ID]
            if not self.checkValueAgainstStats(widget, supplyStats):
                return False
            
            # https://forums.frontier.co.uk/showthread.php?t=34986&p=1162429&viewfull=1#post1162429
            # It seems that stations usually pay within 25% of the
            # asking price as a buy-back price. If the user gives
            # us something out of those bounds, check with them.
            if paying < int(asking * 0.70) or \
                    paying < asking - 250:
                widget.bell()
                ok = mbox.askokcancel(
                        "Are you sure about that?",
                        "You've indicated that the station is:\n"
                        "  PAYING: {:>10n}cr\n"
                        "  ASKING: {:>10n}cr\n"
                        "for {}.\n"
                        "\n"
                        "This is outside of expected tolerances, "
                        "please check the numbers.\n"
                        "\n"
                        "If the numbers are correct, click OK and "
                        "please post a screenshot of the market UI "
                        "to the TD thread "
                        "(http://kfs.org/td/thread)."
                        .format(
                                paying, asking,
                                widget.item.name
                                ),
                        default=mbox.CANCEL,
                        parent=widget,
                        )
                if not ok:
                    self.selectWidget(widget, "")
                    return False
            
            return True
        
        if pos == 3:
            value = widget.val.get()
            if value == "0":
                value = "-"
            if value == "":
                value = "?"
            re.sub(",", "", value)
            widget.val.set(value)
            if re.match(r'^(-|\?|\d+[LlMmHh\?])$', value):
                return True
            mbox.showerror(
                        "Invalid supply value",
                        "If the item is in-supply, this field should "
                        "be the number of units followed by one of 'L', "
                        "'M' or 'H'.\n"
                        "\n"
                        "If the item is out-of-supply, use '-' or '0'.\n"
                        "\n"
                        "EXAMPLE: If the UI says '2341 LOW', type '2341L'.\n"
                    )
            widget.val.set("?")
            self.selectWidget(widget)
            return False
        
        return True


    def onShiftTab(self, event):
        """ Process user pressing Shift+TAB """
        
        widget = event.widget
        if not self.validate(widget):
            return "break"
        if widget.pos > 0 or widget.item.displayNo > 0:
            # Natural flow
            return
        
        self.parent.bell()
        return "break"


    def onTab(self, event):
        """ Process user pressing TAB """
        
        widget = event.widget
        if not self.validate(widget):
            return "break"
        if widget.pos + 1 < len(widget.row):
            return
        if widget.item.displayNo + 1 < len(self.itemDisplays):
            return
        self.parent.bell()
        return "break"


    def onReturn(self, event):
        """
            When the user hits <Return>, advance to
            the first cell on the next line, or if
            we are at the bottom of the list, beep.
        """
        
        widget = event.widget
        if not self.validate(widget):
            return "break"
        # advance to the first entry on the next row
        newDisplayNo = widget.item.displayNo + 1
        if newDisplayNo >= len(self.itemDisplays):
            self.parent.bell()
            return "break"
        
        self.focusOn(newDisplayNo, 0)


    def onUp(self, event):
        """ Handle the user pressing up, go up a row """
        
        widget = event.widget
        if not self.validate(widget):
            return "break"
        # can we go up a line?
        if widget.item.displayNo <= 0:
            self.parent.bell()
            return "break"
        
        self.focusOn(widget.item.displayNo - 1, widget.pos)


    def onDown(self, event):
        """ Handle the user pressing down, go down a row """
        
        widget = event.widget
        if not self.validate(widget):
            return "break"
        # can we go up a line?
        newDisplayNo = widget.item.displayNo + 1
        if newDisplayNo >= len(self.itemDisplays):
            self.parent.bell()
            return "break"
        
        self.focusOn(newDisplayNo, widget.pos)


    def endRow(self):
        self.rowNo += 1
        self.colNo = 0


    def addWidget(self, widget, **kwargs):
        widget.grid(row=self.rowNo, column=self.colNo, **kwargs)
        self.colNo += 1
        return widget


    def addLabel(self, text):
        lab = tk.Label(self.interior, text=text)
        return self.addWidget(lab, sticky='W', padx=16)


    def addSection(self, text):
        widget = tk.Label(self.interior, text=text, fg='blue')
        widget.grid(row=self.rowNo, column=0, columnspan=5, sticky='W')
        self.endRow()
        return widget


    def addInput(self, item, defValue, row):
        pos = len(row)
        
        inputVal = tk.StringVar()
        inputVal.set(str(defValue))
        
        inputEl = tk.Entry(self.interior,
                width=10,
                justify=tk.RIGHT,
                textvariable=inputVal)
        inputEl.item = item
        inputEl.row = row
        inputEl.pos = pos
        inputEl.val = inputVal
        
        inputEl.bind('<Shift-Key-Tab>', self.onShiftTab)
        inputEl.bind('<Key-Tab>', self.onTab)
        inputEl.bind('<Key-Return>', self.onReturn)
        inputEl.bind('<Key-Down>', self.onDown)
        inputEl.bind('<Key-Up>', self.onUp)
        inputEl.bind('<FocusOut>', lambda evt: self.validateRow(evt.widget.row))
        self.addWidget(inputEl, sticky='E')
        row.append((inputEl, inputVal))


    def addHeadings(self):
        def addHeading(text):
            self.headings.append(text)
            lab = tk.Label(self.interior, text=text, fg='blue')
            self.addWidget(lab, sticky='W', padx=16)
        addHeading("Item")
        addHeading("Paying")
        addHeading("Asking")
        addHeading("Demand")
        addHeading("Supply")
        addHeading("AvgPay")
        addHeading("AvgAsk")
        self.endRow()


    def addItemRow(self, ID, catID, itemName, paying, asking, demand, supply):
        row = []
        
        displayNo = len(self.itemDisplays)
        item = Item(ID, catID, itemName, displayNo)
        self.items[itemName] = item
        self.itemList.append(item)
        
        demandStats = self.demandStats[item.ID]
        supplyStats = self.supplyStats[item.ID]
        
        self.addLabel(item.name.upper())
        self.addInput(item, paying, row)
        self.addInput(item, asking, row)
        self.addInput(item, demand, row)
        self.addInput(item, supply, row)
        self.addLabel(demandStats[1])
        self.addLabel(supplyStats[1])
        
        self.itemDisplays.append(row)
        
        self.endRow()


    def createWidgets(self):
        self.addHeadings()
        
        tdb, tdenv = self.tdb, self.tdenv
        station = tdenv.startStation
        self.parent.title(station.name())
        
        db = tdb.getDB()
        db.row_factory = sqlite3.Row
        cur = db.cursor()
        
        self.categories = self.tdb.categoryByID
        
        # Do we have entries for this station?
        cur.execute("""
            SELECT  COUNT(*)
              FROM  StationItem
             WHERE  station_id = ?
        """, [station.ID])
        self.newStation = False if int(cur.fetchone()[0]) else True
        
        cur.execute("""
            SELECT  item.item_id,
                    IFNULL(MIN(demand_price), 0),
                    IFNULL(AVG(demand_price), 0),
                    IFNULL(MAX(demand_price), 0)
              FROM  Item
                    LEFT OUTER JOIN StationItem AS si
                        ON (
                            item.item_id = si.item_id
                            AND si.demand_price > 0
                        )
             GROUP  BY 1
        """)
        self.demandStats = {
            ID: [ minCr, int(avgCr), maxCr ]
            for ID, minCr, avgCr, maxCr in cur
        }
        cur.execute("""
            SELECT  item.item_id,
                    IFNULL(MIN(supply_price), 0),
                    IFNULL(AVG(supply_price), 0),
                    IFNULL(MAX(supply_price), 0)
              FROM  Item
                    LEFT OUTER JOIN StationItem AS si
                        ON (
                            item.item_id = si.item_id
                            AND si.supply_price > 0
                        )
             GROUP  BY 1
        """)
        self.supplyStats = {
            ID: [ minCr, int(avgCr), maxCr ]
            for ID, minCr, avgCr, maxCr in cur
        }
        
        if self.newStation and not tdenv.all:
            def splashScreen():
                mbox.showinfo(
                        "New Station!",
                        "There is currently no price data for this station "
                        "in the database!\n\n" + updateUiHelp,
                        parent=self,
                        )
                self.itemDisplays[0][0][0].focus_set()
            self.parent.after(750, splashScreen)
        
        # if the user specified "--all", force listing of all items
        fetchAll = (self.newStation or tdenv.all)
        siJoin = "LEFT OUTER" if fetchAll else "INNER"
        stmt = """
            SELECT  item.category_id AS catID,
                    item.item_id AS ID,
                    item.name AS name,
                    IFNULL(si.demand_price, '') AS paying,
                    IFNULL(si.supply_price, '') AS asking,
                    IFNULL(si.demand_units, 0) AS demandUnits,
                    IFNULL(si.demand_level, 0) AS demandLevel,
                    IFNULL(si.supply_units, 0) AS supplyUnits,
                    IFNULL(si.supply_level, 0) AS supplyLevel
              FROM  (
                        Category AS cat
                            INNER JOIN Item item
                                USING (category_id)
                    ) {siJoin} JOIN
                        StationItem si ON (
                            si.item_id = item.item_id AND si.station_id = ?
                        )
             ORDER  BY cat.name, item.ui_order
        """.format(siJoin=siJoin)
        tdenv.DEBUG1("sql: {}; bind: {}", stmt, station.ID)
        cur.execute(stmt, [station.ID])
        
        def describeSupply(units, level):
            if not level:
                return ""
            if units == -1 and level == -1:
                return "?"
            if level == -1:
                return "{}?".format(units)
            return "{}{}".format(units, ['L', 'M', 'H'][level-1])
        
        lastCat = None
        for row in cur:
            cat = row["catID"]
            if cat != lastCat:
                self.addSection(self.categories[cat].name())
                lastCat = cat
            itemName = row["name"]
            paying, asking = row["paying"], row["asking"]
            demand = describeSupply(row["demandUnits"], row["demandLevel"])
            supply = describeSupply(row["supplyUnits"], row["supplyLevel"])
            self.addItemRow(row["ID"], cat, itemName, paying, asking, demand, supply)


        for row in self.itemDisplays:
            self.validateRow(row)


    def getResults(self):
        lastCat = None
        
        txt = (
                "# Generated by TDGUI\n"
                "\n"
                "@ {stn}\n".format(
                    stn=self.tdenv.startStation.name(),
                )
            )
        for item in self.itemList:
            if item.catID != lastCat:
                lastCat = item.catID
                txt += (" + {}\n".format(self.categories[lastCat].dbname))
            
            row = self.itemDisplays[item.displayNo]
            rowvals = [ val[1].get() for val in row ]
            
            itemName = item.name
            paying = int(rowvals[0] or 0)
            asking = int(rowvals[1] or 0)
            demand = rowvals[2]
            supply = rowvals[3]
            
            if not paying and not asking:
                demand, supply = "-", "-"
            else:
                if paying and not demand:
                    demand = "?"
                
                if asking == 0:
                    supply = "-"
                elif not supply:
                    supply = "?"
                elif re.match(r"^\d+$", supply):
                    if int(supply) != 0:
                        supply += '?'
            
            txt += ("     {item:<30s} "
                    "{paying:>10} "
                    "{asking:>10} "
                    "{demand:>10} "
                    "{supply:>10}\n".format(
                        item=itemName,
                        paying=paying,
                        asking=asking,
                        demand=demand,
                        supply=supply
                        ))
        self.results = txt


def render(tdb, tdenv, tmpPath):
    parent = tk.Tk()
    gui = UpdateGUI(parent, tdb, tdenv)
    gui.mainloop()
    gui.getResults()
    with tmpPath.open("w", encoding="utf-8") as fh:
        print(gui.results, file=fh)
